from django.conf import settings

from meta.views import MetadataMixin

from ...core.mixins import WebpackBuiltTemplateViewMixin
from ... import app_settings


class HomeView(MetadataMixin, WebpackBuiltTemplateViewMixin):
    description = settings.META_DESCRIPTION
    url = app_settings.APP_URL



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import ProfileApiViewSet, ProfileDetailApiView


_base_urlpatterns = [
    url(r'^$', ProfileApiViewSet.as_view({'get': 'list'}), name='rest-list'),

    url(r'^(?P<sso_id>[-_\w]+)/$', ProfileApiViewSet.as_view({'get': 'retrieve', 'patch': 'partial_update'}),
        name="rest-detail"),
]

_urlpatterns = _base_urlpatterns

extra_urlpatterns = (format_suffix_patterns([
                         url(r'^profile/$', ProfileDetailApiView.as_view(), name="detail")]),
                     'django_sso_app_profile_extra')

base_urlpatterns = (format_suffix_patterns(_base_urlpatterns), 'django_sso_app_profile')
urlpatterns = (format_suffix_patterns(_urlpatterns), 'django_sso_app_profile')



import json
import jwt

from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.core import mail

from allauth.account.adapter import get_adapter
from rest_framework import status

from django_sso_app.core.tests.factories import UserTestCase

from .... import app_settings

User = get_user_model()


class TestUserCreate(UserTestCase):

    def test_staff_user_can_create_user_with_profile_with_unverified_email(self):
        new_staff_user = self._get_new_staff_user()
        new_user_object = self._get_new_user_object()

        new_staff_user_token_headers = self._get_new_api_token_headers(new_staff_user)

        client = self._get_client()
        response = client.post(
            reverse('django_sso_app_user:rest-list'),
            data=json.dumps(new_user_object),
            **new_staff_user_token_headers
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_user_sso_id = response.data.get('sso_id')

        response2 = client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user_sso_id,)),
            **new_staff_user_token_headers
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data['email'], new_user_object['email'])
        self.assertEqual(response2.data['email_verified'], False)

    def test_staff_user_can_create_user_with_profile_and_validated_email(self):
        new_staff_user = self._get_new_staff_user()
        new_user_object = self._get_new_user_object()

        new_staff_user_token_headers = self._get_new_api_token_headers(new_staff_user)

        client = self._get_client()
        response = client.post(
            reverse('django_sso_app_user:rest-list') + '?skip_confirmation=true',
            data=json.dumps(new_user_object),
            **new_staff_user_token_headers
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_user_sso_id = response.data.get('sso_id')

        response2 = client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user_sso_id,)),
            **new_staff_user_token_headers
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data['email'], new_user_object['email'])
        self.assertEqual(response2.data['email_verified'], True)

    def test_non_staff_user_can_not_create_profile(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        new_user_object = self._get_new_user_object()

        response = client.post(
            reverse('django_sso_app_user:rest-list') + '?skip_confirmation=true',
            data=json.dumps(new_user_object),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST, 'non staff users can create new user')


class TestUserRetrieve(UserTestCase):

    def test_user_can_retrieve_user(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user.sso_app_profile.sso_id,)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

    def test_staff_user_can_retrieve_user_with_token(self):
        new_user = self._get_new_user()
        new_staff_user = self._get_new_staff_user()

        new_staff_user_token_headers = self._get_new_api_token_headers(new_staff_user)

        client = self._get_client()
        response = client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user.sso_app_profile.sso_id,)),
            **new_staff_user_token_headers
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)


class TestUserUpdate(UserTestCase):

    def test_user_email_update_deletes_all_devices_and_waits_for_confirmation(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)
        new_user_profile = new_user.sso_app_profile

        first_rev = new_user_profile.sso_rev

        first_email_login = self._get_login_object(new_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(first_email_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        new_email = self._get_random_email()

        response2 = client.patch(
            reverse('django_sso_app_user:rest-detail', args=[new_user_profile.sso_id]),
            data=json.dumps({
                'email': new_email
            }),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        emails = new_user.emailaddress_set.all()
        self.assertEqual(emails.count(), 2, 'email not created')

        self.assertEqual(len(mail.outbox), 1, 'no email sent')

        cofirmation_message = mail.outbox[0]
        recipients = cofirmation_message.recipients()

        self.assertIn(new_email, recipients, 'user email not in recipients')

        client = self._get_client()
        response3 = client.post(
            reverse('rest_login'),
            data=json.dumps(first_email_login),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, 200, 'unable to login with old email')

        client = self._get_client()
        response4 = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response4.status_code, 400, 'login with unconfirmed email is still possible')

        unconfirmed_email = new_user.emailaddress_set.get(email=new_email)
        self.assertEqual(unconfirmed_email.verified, False)

        adapter = get_adapter()
        adapter.confirm_email(None, unconfirmed_email)

        client = self._get_client()
        response5 = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response5.status_code, 200, 'login with confirmed email is not possible')

        self.assertEqual(new_user.emailaddress_set.count(), 1, 'more than 1 email after new email confirmation')

        client = self._get_client()
        response6 = client.post(
            reverse('rest_login'),
            data=json.dumps(first_email_login),
            content_type='application/json'
        )

        self.assertEqual(response6.status_code, 400, 'login old email is still possible')

    def test_user_password_update_refreshes_and_returns_current_device_jwt_and_deletes_all_profile_devices(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)
        new_user_profile = new_user.sso_app_profile

        first_rev = new_user_profile.sso_rev

        first_email_login = self._get_login_object(new_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(first_email_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        jwt_cookie = self._get_response_jwt_cookie(response)

        self.assertIsNotNone(jwt_cookie, 'no jwt cookie set')

        unverified_payload = jwt.decode(jwt_cookie, None, False)

        updated_pass = self._get_random_pass()

        response2 = client.patch(
            reverse('django_sso_app_user:rest-detail', args=[new_user_profile.sso_id]),
            data=json.dumps({
                'password': updated_pass
            }),
            content_type='application/json'
        )

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        jwt_cookie2 = self._get_response_jwt_cookie(response2)

        self.assertIsNotNone(jwt_cookie2, 'no jwt cookie set')
        self.assertEqual(new_user_profile.devices.count(), 0, 'other devices not deleted')

        client2 = self._get_client()
        response3 = client2.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, updated_pass)),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, 200, 'unable to login with new password')

        jwt_cookie3 = self._get_response_jwt_cookie(response3)
        unverified_payload3 = jwt.decode(jwt_cookie3, None, False)

        self.assertGreater(unverified_payload3.get('sso_rev'), unverified_payload.get('sso_rev'),
                           'rev not incremented after pass update')

    def test_user_password_update_refreshes_profile_even_if_called_with_authorization_headers(self):
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)
        new_user_profile = new_user.sso_app_profile

        first_rev = new_user_profile.sso_rev

        first_email_login = self._get_login_object(new_user.email, new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(first_email_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        received_jwt = self._get_response_jwt_cookie(response)
        unverified_payload = jwt.decode(received_jwt, None, False)

        self.assertEqual(response.data['token'], received_jwt, 'jwt differs between cookie and response')

        updated_pass = self._get_random_pass()
        valid_password_update = {
            'password': updated_pass
        }

        response2 = self._get_client().patch(
            reverse('django_sso_app_user:rest-detail', args=[new_user_profile.sso_id]),
            data=json.dumps(valid_password_update),
            content_type='application/json',
            HTTP_AUTHORIZATION='Bearer {}'.format(received_jwt)
        )

        self.assertEqual(response2.status_code, 200, 'cannot patch with jwt header')

        received_jwt2 = self._get_response_jwt_cookie(response2)

        self.assertIsNotNone(received_jwt2, 'no jwt cookie received')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        devices = new_user_profile.devices.all()

        self.assertEqual(devices.count(), 0, 'devices not deleted')

        response3 = self._get_client().post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, updated_pass)),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, 200, 'unable to login with new password')

        jwt_cookie3 = self._get_response_jwt_cookie(response3)
        unverified_payload3 = jwt.decode(jwt_cookie3, None, False)

        self.assertGreater(unverified_payload3.get('sso_rev'), unverified_payload.get('sso_rev'),
                           'rev not increased after pass update')

    def test_user_email_update_by_staff_disables_user_login_user_must_confirm_email(self):
        """
        User update by staff refreshes disables login and deletes all user devices
        """
        new_email = self._get_random_email()
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(email=new_email, password=new_pass)
        new_user_profile = new_user.sso_app_profile

        new_staff_user = self._get_new_staff_user()

        first_rev = new_user_profile.sso_rev

        updated_email = self._get_random_email()

        response = self._get_client().patch(
            reverse('django_sso_app_user:rest-detail', args=[new_user_profile.sso_id]),
            data=json.dumps({
                'email': updated_email
            }),
            **self._get_new_api_token_headers(new_staff_user)
        )

        received_jwt = self._get_response_jwt_cookie(response)

        self.assertIsNone(received_jwt, 'jwt cookie received')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertEqual(new_user_profile.devices.count(), 0, 'user devices not deleted')

        response3 = self._get_client().post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(updated_email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, status.HTTP_400_BAD_REQUEST,
                         'user can login without confirming email after update')

        msg = _('We have sent an e-mail to you for verification. '
                'Follow the link provided to finalize the signup process. '
                'Please contact us if you do not receive it within a few minutes.')
        self.assertEqual(response3.data, msg)

        response4 = self._get_client().post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response4.status_code, status.HTTP_200_OK, 'user can not login with old confirmed email')

    def test_user_email_update_by_staff_with_skip_email_confirmation_activates_new_email(self):
        """
        User update by staff refreshes disables login and deletes all user devices
        """
        new_email = self._get_random_email()
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(email=new_email, password=new_pass)
        new_user_profile = new_user.sso_app_profile

        new_staff_user = self._get_new_staff_user()

        first_rev = new_user_profile.sso_rev

        updated_email = self._get_random_email()

        response = self._get_client().patch(
            reverse('django_sso_app_user:rest-detail', args=[new_user_profile.sso_id]) + '?skip_confirmation=true',
            data=json.dumps({
                'email': updated_email
            }),
            **self._get_new_api_token_headers(new_staff_user)
        )

        received_jwt = self._get_response_jwt_cookie(response)

        self.assertIsNone(received_jwt, 'jwt cookie received')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertEqual(new_user_profile.devices.count(), 0, 'user devices not deleted')

        response2 = self._get_client().post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(updated_email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK, 'user can not login with old confirmed email')



import os

import cherrypy

from django_sso_app.config.wsgi import application

SOCKET_FILE = '/tmp/app.sock'
CHERRYPY_THREAD_POOL_SIZE = os.environ.get('CHERRYPY_THREAD_POOL_SIZE', 30)
CHERRYPY_SOCKET_HOST = os.environ.get('CHERRYPY_SOCKET_HOST', '0.0.0.0')
CHERRYPY_SOCKET_PORT = os.environ.get('CHERRYPY_SOCKET_PORT', 5000)


def run(host=CHERRYPY_SOCKET_HOST, port=CHERRYPY_SOCKET_PORT):
    # Cleanup socket file
    try:
        os.remove(SOCKET_FILE)
    except OSError:
        pass

    # Mount the application
    cherrypy.tree.graft(application, "/")

    # Unsubscribe the default server
    cherrypy.server.unsubscribe()

    # Instantiate a new server object
    server = cherrypy._cpserver.Server()

    # Configure the server object
    server.socket_host = host
    server.socket_port = port
    server.thread_pool = CHERRYPY_THREAD_POOL_SIZE

    # If in same node use socket files
    # https://lists.freebsd.org/pipermail/freebsd-performance/2005-February/001143.html
    server.socket_file = SOCKET_FILE

    # For SSL Support
    # server.ssl_module            = 'pyopenssl'
    # server.ssl_certificate       = 'ssl/certificate.crt'
    # server.ssl_private_key       = 'ssl/private.key'
    # server.ssl_certificate_chain = 'ssl/bundle.crt'

    # Subscribe this server
    server.subscribe()

    # Example for a 2nd server (same steps as above):
    # Remember to use a different port

    # server2             = cherrypy._cpserver.Server()

    # server2.socket_host = "0.0.0.0"
    # server2.socket_port = 8081
    # server2.thread_pool = 30
    # server2.subscribe()

    # Start the server engine (Option 1 *and* 2)

    cherrypy.engine.start()

    # change socket permissions
    os.chmod(SOCKET_FILE, 0o777)

    cherrypy.engine.block()


if __name__ == '__main__':
    run()



"""
django envs:
    SESSION_ENGINE, SESSION_COOKIE_NAME, SESSION_COOKIE_PATH, \
    SESSION_COOKIE_DOMAIN, SESSION_SAVE_EVERY_REQUEST, SESSION_COOKIE_SECURE, SESSION_COOKIE_HTTPONLY, \
    SESSION_COOKIE_SAMESITE
"""

import environ

env = environ.Env()

# common
DEBUG = env.bool("DJANGO_DEBUG", default=True)
APP_DOMAIN = env("APP_DOMAIN", default="localhost:8000")
COOKIE_DOMAIN = env("COOKIE_DOMAIN", default=APP_DOMAIN.split(':')[0])
ACCOUNT_DEFAULT_HTTP_PROTOCOL = env("ACCOUNT_DEFAULT_HTTP_PROTOCOL", default='http' if DEBUG else 'https')
I18N_PATH_ENABLED = env.bool('I18N_PATH_ENABLED', default=False)

REDIS_ENABLED = env.bool('REDIS_ENABLED', default=False)

# django-sso-app
DJANGO_SSO_APP_SOCIALACCOUNT_PROVIDERS = env.list('DJANGO_SSO_APP_SOCIALACCOUNT_PROVIDERS', default=[])

BACKEND_SHAPES = ('backend_standalone',
                  'backend_only',
                  'backend_only_apigateway',
                  'backend_app',
                  'backend_app_apigateway')
APP_SHAPES = ('app',
              'app_persistence',
              'app_apigateway',
              'app_persistence_apigateway')

AVAILABLE_SHAPES = BACKEND_SHAPES + APP_SHAPES

DJANGO_SSO_APP_SHAPE = env('DJANGO_SSO_APP_SHAPE', default='backend_only')

DJANGO_SSO_APP_BASE_DJANGO_APPS = [
    'django_sso_app',

    'django_sso_app.core.apps.status',
    'django_sso_app.core.apps.profiles',
    'django_sso_app.core.apps.groups',
    'django_sso_app.core.apps.devices',
    'django_sso_app.core.apps.services',
    'django_sso_app.core.apps.api_gateway',
    'django_sso_app.core.apps.passepartout'
]

DJANGO_SSO_APP_DJANGO_APPS = [
    'allauth',
    'allauth.account',
    'allauth.socialaccount',

    'rest_framework.authtoken',

    'django_countries',

    'treebeard'

] + DJANGO_SSO_APP_BASE_DJANGO_APPS

if 'google' in DJANGO_SSO_APP_SOCIALACCOUNT_PROVIDERS:
    DJANGO_SSO_APP_DJANGO_APPS += [
        'allauth.socialaccount.providers.google',
    ]

# django
# SESSION_ENGINE = 'django.contrib.sessions.backends.cache'

COUNTRIES_FIRST = [
    'IT',
]



from django.contrib.auth import get_user_model, forms
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

User = get_user_model()


class UserChangeForm(forms.UserChangeForm):
    class Meta(forms.UserChangeForm.Meta):
        model = User


class UserCreationForm(forms.UserCreationForm):

    error_message = forms.UserCreationForm.error_messages.update(
        {"duplicate_username": _("This username has already been taken.")}
    )

    class Meta(forms.UserCreationForm.Meta):
        model = User

    def clean_username(self):
        username = self.cleaned_data["username"]

        try:
            User.objects.get(username=username)
        except User.DoesNotExist:
            return username

        raise ValidationError(self.error_messages["duplicate_username"])

    # pai
    def __init__(self, *args, **kwargs):
        super(UserCreationForm, self).__init__(*args, **kwargs)

        # removing unnecessary password fields
        self.fields['password1'].required = False
        self.fields['password2'].required = False
        # If one field gets autocompleted but not the other, our 'neither
        # password or both password' validation will be triggered.
        self.fields['password1'].widget.attrs['autocomplete'] = 'off'
        self.fields['password2'].widget.attrs['autocomplete'] = 'off'



# Boostable functions
#
# import pyximport
# pyximport.install()


def get_apigateway_profile_groups_from_header(groups_header: str) -> list:
    """
    Extracts apigateway consumer groups from header
    :param groups_header:
    :return:
    """
    if groups_header not in (None, ''):
        return list(map(str.strip, groups_header.split(',')))

    return []



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import passepartout_login_view

_urlpatterns = [
    url(r'^login/(?P<token>\w+)/$',
        passepartout_login_view,
        name='login')
]

urlpatterns = (format_suffix_patterns(_urlpatterns), 'django_sso_app_passepartout')



import logging
from urllib.parse import urlencode, urljoin

# import pyximport
# pyximport.install()

from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse

from allauth.account.adapter import get_adapter

from ...utils import invalidate_cookie, set_cookie, set_session_key, get_session_key
from ...permissions import is_authenticated
from ...functions import get_url_host
from ... import app_settings
from ..services.models import Service
from .models import Passepartout
from .functions import get_referer

logger = logging.getLogger('django_sso_app.core.apps.passepartout')


# @csrf_exempt
def passepartout_login_view(request, token):
    """
    Receives unguessable token by path parameter and destination url by get parameter.
    1) checks user subscription to destination service (obtained by next param as Service url)
    2) checks its position in DJANGO_SSO_APP_URLS ordered dictionary
    3) redirects to next APP_URL if there are subsquent sso domains or directly to service url (if redirecting to APP_URL request will include a jwt token COOKIES)
    """
    try:
        nextUrl = request.GET.get('next', None)
        logger.debug('ENTERING passepartout with next {}'.format(nextUrl))

        if nextUrl is None:
            raise Http404("No next url")

        try:
            passepartout = Passepartout.objects.get(is_active=True, token=token)

        except Passepartout.DoesNotExist:
            logger.warning("Passepartout token not found, redirecting to login")

            response = redirect(reverse('account_login'))
            # invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)

            return response

        referer = request.META.get('HTTP_REFERER', None)
        if referer is None:
            referer = get_referer(app_settings.APP_DOMAIN,
                                  app_settings.BACKEND_DOMAINS_DICT,
                                  app_settings.BACKEND_DOMAINS,
                                  app_settings.BACKEND_FULL_URLS_CHAIN)

        comes_from = get_url_host(referer)
        sso_urls_chain = app_settings.BACKEND_URLS_CHAIN
        sso_urls_chain_dict = app_settings.BACKEND_URLS_CHAIN_DICT
        logger.debug('Passepartout chain dict: {0}, this is: {1}. Comes from {2} with nextUrl: {3}'.format(sso_urls_chain,
                                                                                                           app_settings.APP_DOMAIN,
                                                                                                           comes_from,
                                                                                                           nextUrl))

        previous_sso_instance_index = sso_urls_chain_dict[comes_from]
        last_sso_instance_index = len(sso_urls_chain) - 1
        next_sso_instance_index = last_sso_instance_index + 1

        is_last_bump = previous_sso_instance_index == last_sso_instance_index
        redirect_args = None

        next_url_base_path = get_url_host(nextUrl)

        if next_url_base_path is None:
            if is_last_bump:
                next_url_base_path = sso_urls_chain[last_sso_instance_index]
            else:
                next_url_base_path = get_url_host(referer)

            oldNextUrl = nextUrl
            nextUrl = urljoin(next_url_base_path, nextUrl)

            logger.debug('next url "{}" has no base_path, setting "{}"'.format(oldNextUrl, nextUrl))

        if next_url_base_path not in app_settings.BACKEND_FULL_URLS_CHAIN:
            try:
                _service = Service.objects.get(service_url=next_url_base_path)

            except Service.DoesNotExist:
                raise Http404("Service not found")

        set_session_key(request, '__dssoa__device__fingerprint', passepartout.device.fingerprint)

        if is_last_bump:
            logger.info('Is last login bump')
            passepartout.is_active = False
            passepartout.save()
            goes_to = nextUrl
        else:
            redirect_args = {
                'next': nextUrl
            }
            goes_to = sso_urls_chain[next_sso_instance_index]

        # if subscribed is None:
        #    subscrition_created = user.subscribe_to_service(service)
        #    logger.info('Did passepartout subscribed user, {0} to {2}? {3}'.format(user, nextUrl, subscription_created))
        #    redirect_args['subscribed'] = 'true'

        formatted_redirect_args = ''
        if not is_last_bump:
            formatted_redirect_args = '?' + urlencode(redirect_args)

        redirect_to = goes_to + formatted_redirect_args
        logger.info('Passepartout login redirecting to {0}'.format(redirect_to))

        response = redirect(redirect_to)

        user = passepartout.device.profile.user

        adapter = get_adapter(request)

        try:
            adapter.login(request, user)

        except Exception as e:
            logger.exception('adapter login error {}, invalidating cookie'.format(e))

            invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)  # invalidate cookie

        else:
            if is_authenticated(user):
                jwt_token = get_session_key(request, '__dssoa__jwt_token', None)

                if jwt_token is None:
                    logger.error('No jwt_token in session')
                    raise Exception('No jwt_token in session')
                else:
                    set_cookie(response, app_settings.JWT_COOKIE_NAME, jwt_token)

            else:
                logger.warning('user "{}" can not authenticate'.format(user))

    except Exception as e:
        logger.error('Passepartout login ERROR: {}'.format(e), exc_info=True)
        raise

    return response



# Generated by Django 2.2.11 on 2020-04-17 15:57

from django.db import migrations, models
import django_sso_app.core.apps.profiles.models

def forwards_func(apps, schema_editor):
    # We get the model from the versioned app registry;
    # if we directly import it, it'll be the wrong version
    Profile = apps.get_model("django_sso_app", "Profile")
    db_alias = schema_editor.connection.alias
    for p in Profile.objects.using(db_alias):
        p.django_user_email = p.user.email
        p.django_user_username = p.user.username
        p.save()


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0003_emailaddress'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='profile',
            managers=[
                ('objects', django_sso_app.core.apps.profiles.models.SsoIdPKManager()),
            ],
        ),
        migrations.AddField(
            model_name='profile',
            name='django_user_email',
            field=models.EmailField(blank=True, editable=False, max_length=254, null=True, verbose_name='email address'),
        ),
        migrations.AddField(
            model_name='profile',
            name='django_user_username',
            field=models.CharField(blank=True, editable=False, max_length=150, null=True, verbose_name='user username'),
        ),
        migrations.RunPython(forwards_func, reverse_func),
    ]



import logging
# import inspect
# import warnings

from django.db import transaction
from django.urls import reverse
from django.contrib.auth import authenticate
from django.utils.translation import activate as activate_translation

from django.conf import settings
from django.utils.translation import gettext_lazy as _

from rest_framework.renderers import JSONRenderer

from allauth.account.adapter import DefaultAccountAdapter, app_settings as allauth_app_settings

from .apps.emails.utils import unconfirm_all_user_emails
from .apps.passepartout.utils import get_passepartout_login_redirect_url
from . import app_settings

logger = logging.getLogger('django_sso_app.core')


class AccountAdapter(DefaultAccountAdapter):
    def __init__(self, request=None):
        super(AccountAdapter, self).__init__(request)

        self.error_messages = DefaultAccountAdapter.error_messages
        self.error_messages['email_not_verified'] = _('We have sent an e-mail to you for verification. Follow the link provided to finalize the signup process. Please contact us if you do not receive it within a few minutes.')

    def populate_username(self, request, user):
        # new user has username == email
        user.username = user.email

    def is_ajax(self, request):
        is_ajax = request.path.startswith('/api') or \
                  super(AccountAdapter, self).is_ajax(request) or \
                  request.META.get('CONTENT_TYPE', None) == 'application/json'

        return is_ajax

    def ajax_response(self, request, response, redirect_to=None, form=None, data=None):
        # manually setting response renderer
        setattr(response, 'accepted_renderer', JSONRenderer())
        setattr(response, 'accepted_media_type', 'application/json')
        setattr(response, 'renderer_context', {})

        return super(AccountAdapter, self).ajax_response(request, response, redirect_to, form, data)

    def get_login_redirect_url(self, request):
        """
        Returns the default URL to redirect to after logging in.  Note
        that URLs passed explicitly (e.g. by passing along a `next`
        GET parameter) take precedence over the value returned here.
        """

        if request.path.startswith(reverse('account_signup')) or request.path.startswith(reverse('rest_signup')):
            logger.info('do not get redirect_url from signup')
            redirect_url = None
        else:
            redirect_url = get_passepartout_login_redirect_url(request)

            if redirect_url is None:
                redirect_url = super(AccountAdapter, self).get_login_redirect_url(request)

        return redirect_url

    def login(self, request, user):
        # HACK: This is not nice. The proper Django way is to use an
        # authentication backend
        logger.debug('Adapter login')

        super(AccountAdapter, self).login(request, user)

    def logout(self, request):
        # allauth logout
        super(AccountAdapter, self).logout(request)

    def confirm_email(self, request, email_address):
        """
        Marks the email address as confirmed on the db,
        deletes old user emails.
        """
        email_address.verified = True
        email_address.set_as_primary(conditional=False)
        email_address.save()

        # one email per user
        email_address.user.emailaddress_set.exclude(email=email_address).delete()

    def set_password(self, user, password):
        user.set_password(password)
        user.save()

    @transaction.atomic
    def save_user(self, request, user, form, commit=True):
        """
        Saves a new User instance using information provided in the
        signup form.
        """
        return super(AccountAdapter, self).save_user(request, user, form, commit)

    def render_mail(self, template_prefix, email, context):
        user = context.get('user')
        logger.info('render mail for "{}"'.format(user))
        user_language = user.sso_app_profile.language

        activate_translation(user_language)
        context.update({
            'EMAILS_DOMAIN': settings.EMAILS_DOMAIN,
            'EMAILS_SITE_NAME': settings.EMAILS_SITE_NAME
        })

        return super(AccountAdapter, self).render_mail(template_prefix, email, context)

    def get_email_confirmation_url(self, request, emailconfirmation):
        """Constructs the email confirmation (activation) url.
        Note that if you have architected your system such that email
        confirmations are sent outside of the request context request
        can be None here.
        """
        _url = reverse(
            "account_confirm_email",
            args=[emailconfirmation.key])

        email_confirmation_url = '{}://{}{}'.format(app_settings.HTTP_PROTOCOL,
                                                    settings.EMAILS_DOMAIN,
                                                    _url)

        return email_confirmation_url

    # pai

    def unconfirm_all_user_emails(self, user):
        return unconfirm_all_user_emails(user, self.request)

    def update_user_email(self, user, new_email):
        user.email = new_email  # set user model email

        setattr(user, '__dssoa__email_updated', True)

        user.save()

    def update_user_username(self, user, new_username):
        user.username = new_username  # set user model username

        setattr(user, '__dssoa__username_updated', True)

        user.save()



import logging

from django.db import models
from django.contrib.auth.models import Group as DjangoGroup
from django.utils.translation import ugettext_lazy as _
from django.urls import reverse

from treebeard.mp_tree import MP_Node, MP_NodeManager

from ...models import CreatedAtModel

logger = logging.getLogger('django_sso_app.core.apps.groups')


class Group(DjangoGroup):
    class Meta:
        proxy = True
        app_label = 'django_sso_app'
        verbose_name = _('Group')

    def get_relative_rest_url(self):
        return reverse("django_sso_app_group:rest-detail", args=[self.pk])


class GroupingManager(MP_NodeManager):
    pass


class Grouping(MP_Node, CreatedAtModel):
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Grouping')

    name = models.CharField(_("name"), max_length=150)
    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    type = models.CharField(_("type"), max_length=255, null=True, blank=True)
    description = models.TextField(_('description'), null=True, blank=True)

    objects = GroupingManager()
    node_order_by = ('name', )  # treebeard mp_node ordering

    def get_relative_rest_url(self):
        return reverse("django_sso_app_group:rest-grouping-detail", args=[self.pk])



from ..core.settings.common import *


# django-sso-app
DJANGO_SSO_APP_APIGATEWAY_HOST = env('DJANGO_SSO_APP_APIGATEWAY_HOST', default='http://kong')

DJANGO_SSO_APP_DJANGO_AUTHENTICATION_BACKENDS = [
    # login
    # 'allauth.account.auth_backends.AuthenticationBackend',  # overriden
    'django_sso_app.core.authentication.backends.login.DjangoSsoAppLoginAuthenticationBackend',

    # requests
    'django_sso_app.core.authentication.backends.DjangoSsoAppApiGatewayAuthenticationBackend',
    'django_sso_app.core.authentication.backends.DjangoSsoAppJwtAuthenticationBackend'
]

DJANGO_SSO_APP_DRF_AUTHENTICATION_CLASSES = ['django_sso_app.core.api.authentication.DjangoSsoApiAuthentication']

DJANGO_SSO_APP_BACKEND_CUSTOM_FRONTEND_APP = env('DJANGO_SSO_APP_BACKEND_CUSTOM_FRONTEND_APP', default=None)
if DJANGO_SSO_APP_BACKEND_CUSTOM_FRONTEND_APP is not None:
    DJANGO_SSO_APP_DJANGO_APPS = [
        DJANGO_SSO_APP_BACKEND_CUSTOM_FRONTEND_APP
    ] + DJANGO_SSO_APP_DJANGO_APPS

DJANGO_SSO_APP_USER_TYPES = env.list('DJANGO_SSO_APP_USER_TYPES', default=[])

# django
SESSION_COOKIE_SAMESITE = None  # Safari 12 and chrome cross site
SESSION_EXPIRE_AT_BROWSER_CLOSE = True

_DJANGO_SSO_APP_APIGATEWAY_ENABLED = DJANGO_SSO_APP_SHAPE.find('apigateway') > -1
if _DJANGO_SSO_APP_APIGATEWAY_ENABLED:
    USE_X_FORWARDED_HOST = True
    SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')


# allauth
# ------------------------------------------------------------------------------
ACCOUNT_ALLOW_REGISTRATION = True

# https://django-allauth.readthedocs.io/en/latest/configuration.html
ACCOUNT_AUTHENTICATION_METHOD = 'username_email'

# https://django-allauth.readthedocs.io/en/latest/configuration.html
# ! ACCOUNT_EMAIL_REQUIRED = 'email' in ACCOUNT_AUTHENTICATION_METHOD
ACCOUNT_EMAIL_REQUIRED = True
# https://django-allauth.readthedocs.io/en/latest/configuration.html
ACCOUNT_EMAIL_VERIFICATION = "mandatory" if ACCOUNT_EMAIL_REQUIRED else "optional"  # "none"

ACCOUNT_ADAPTER = "django_sso_app.core.adapter.AccountAdapter"
ACCOUNT_FORMS = {
    'signup': 'django_sso_app.core.forms.SignupForm',
    'login': 'django_sso_app.core.forms.LoginForm'
}

# https://django-allauth.readthedocs.iobackend/en/latest/configuration.html
# https://django-allauth.readthedocs.io/en/latest/configuration.html
# SOCIALACCOUNT_ADAPTER = "django_sso_app.backend.users.adapters.SocialAccountAdapter"

ACCOUNT_CONFIRM_EMAIL_ON_GET = True

# ! ACCOUNT_USERNAME_REQUIRED = 'username' in DJANGO_SSO_APP_USER_FIELDS
ACCOUNT_USERNAME_REQUIRED = False  # set to email by adapter

ACCOUNT_AUTHENTICATED_LOGIN_REDIRECTS = False
ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = True
ACCOUNT_EMAIL_CONFIRMATION_AUTHENTICATED_REDIRECT_URL = '/'
ACCOUNT_SESSION_REMEMBER = False

# case insensitive username
ACCOUNT_PRESERVE_USERNAME_CASING = True

SOCIALACCOUNT_PROVIDERS = {}

if 'google' in DJANGO_SSO_APP_SOCIALACCOUNT_PROVIDERS:
    SOCIALACCOUNT_PROVIDERS['google'] = {
        'SCOPE': [
            'profile',
            'email',
        ],
        'AUTH_PARAMS': {
            'access_type': 'online',
        }
    }



import logging
import requests

from django.contrib.auth import get_user_model
from django.db import transaction

from .... import app_settings

logger = logging.getLogger('django_sso_app.core.apps.api_gateway.kong')
User = get_user_model()


def create_apigw_consumer(custom_id):
    logger.info('creating apigw consumer with custom_id {}'.format(custom_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/"
    data = {'custom_id': custom_id}

    r = requests.post(url, json=data)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.debug('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.debug('apigw response ({0}) {1}'.format(status_code, response_body))

    return status_code, response_body


def get_apigw_consumer(consumer_id, custom_id=None):
    if custom_id is None:
        logger.info('getting apigw consumer {}'.format(consumer_id))
        url = app_settings.APIGATEWAY_HOST + "/consumers/{}".format(consumer_id)
    else:
        logger.info('getting apigw consumer with custom_id {}'.format(custom_id))
        url = app_settings.APIGATEWAY_HOST + "/consumers/?custom_id={}".format(custom_id)

    r = requests.get(url)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.debug('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.debug('apigw response ({0}) {1}'.format(status_code, response_body))

    if custom_id is None:
        return status_code, response_body
    else:
        if status_code == 200:
            if len(response_body['data']):
                return status_code, response_body['data'][0]
            else:
                return 404, None
        else:
            return status_code, response_body


def delete_apigw_consumer(consumer_id):
    logger.info('deleting apigw consumer for {}'.format(consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id

    r = requests.delete(url)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.debug('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def create_apigw_consumer_jwt(consumer_id):
    logger.info('creating apigw consumer jwt for {}'.format(consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/jwt/"
    data = {}

    r = requests.post(url, json=data)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.debug('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.debug('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def delete_apigw_consumer_jwt(consumer_id, jwt_id):
    logger.info('deleting apigw consumer jwt {} for {}'.format(jwt_id, consumer_id))

    url = '{}/{}/{}/jwt/{}'.format(app_settings.APIGATEWAY_HOST, "consumers", consumer_id, jwt_id)

    logger.info('calling kong url "{}"'.format(url))

    r = requests.delete(url)

    status_code = r.status_code

    logger.info('apigw response ({0}) {1}'.format(status_code, None))
    return r.status_code, None


def get_apigw_consumer_jwts(consumer_id):
    logger.info('getting apigw consumer jwts for {}'.format(consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/jwt/"

    r = requests.get(url)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def create_apigw_consumer_acl(consumer_id, group_name):
    logger.info('creating apigw consumer acl {} for {}'.format(group_name,
                                                               consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/"
    data = {"group": group_name}

    r = requests.post(url, json=data)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def get_apigw_consumer_acls(consumer_id):
    logger.info('getting apigw consumer groups for {}'.format(consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/"

    r = requests.get(url)

    status_code = r.status_code
    try:
        response_body = r.json()
    except:
        logger.info('Response ({}) has no payload'.format(status_code))
        response_body = None

    logger.info('apigw response ({0}) {1}'.format(status_code, response_body))
    return status_code, response_body


def delete_apigw_consumer_acl(consumer_id, group_name):
    logger.info('deleting apigw consumer acl {} for {}'.format(group_name,
                                                               consumer_id))

    url = app_settings.APIGATEWAY_HOST + "/consumers/" + consumer_id + "/acls/" + group_name

    r = requests.delete(url)

    status_code = r.status_code

    logger.info('apigw response ({0}) {1}'.format(status_code, None))
    return r.status_code, None


def add_apigw_consumer_to_groups_acls(consumer_id, group_names):
    # adding apigw consumer to profile groups
    """
    profile_groups = app_settings.DEFAULT_PROFILE_GROUPS + \
                     list(profile.groups.values_list('name', flat=True))
    """

    for group_name in group_names:
        status_code, acl = create_apigw_consumer_acl(consumer_id, group_name)
        if status_code not in (201, 409):  # 409 when already created
            logger.error(
                'Error ({0}) creating apigw consumer acl, {1}'.format(
                    status_code, acl))
            raise Exception(
                "Error {0} creating apigw consumer acl, {1}".format(
                    status_code, acl))


# profile

@transaction.atomic
def get_or_create_profile_apigw_consumer(profile, force_recreate=False):
    logger.info('getting or creating apigw consumer for "{}"'.format(profile))
    consumer_id = profile.apigw_consumer_id

    if consumer_id is None:
        status_code, consumer = get_apigw_consumer(consumer_id, profile.sso_id)
        update_consumer_id = True
    else:
        status_code, consumer = get_apigw_consumer(consumer_id)
        update_consumer_id = False

    if status_code == 200:
        if force_recreate:
            if update_consumer_id:
                consumer_id = consumer['id']

            # deleting old consumer
            _status_code, _res = delete_apigw_consumer(consumer_id)
            # creating new consumer
            _status_code, consumer = create_apigw_consumer(profile.sso_id)

            if status_code == 201:

                # disable sso_rev update
                setattr(profile.user, '__dssoa__apigateway_update', True)
                # setting apigw_consumer_id
                profile.apigw_consumer_id = consumer['id']
                # save profile
                profile.save()
                setattr(profile.user, '__dssoa__apigateway_update', False)

                return consumer
            else:
                logger.error('Error ({0}) creating consumer for "{1}'.format(status_code, profile))
                raise Exception('Error ({0}) creating consumer for "{1}"'.format(status_code, profile))

        else:
            if update_consumer_id:
                # disable sso_rev update
                setattr(profile.user, '__dssoa__apigateway_update', True)
                # setting apigw_consumer_id
                profile.apigw_consumer_id = consumer['id']
                # save profile
                profile.save()
                setattr(profile.user, '__dssoa__apigateway_update', False)

            return consumer

    elif status_code == 404:
        # consumer not found
        status_code, consumer = create_apigw_consumer(profile.sso_id)

        if status_code == 201:
            # disable sso_rev update
            setattr(profile.user, '__dssoa__apigateway_update', True)
            # setting apigw_consumer_id
            profile.apigw_consumer_id = consumer['id']
            # save profile
            profile.save()
            # add consumer to profile groups
            add_apigw_consumer_to_groups_acls(consumer['id'],
                                              app_settings.DEFAULT_PROFILE_GROUPS +
                                              list(profile.groups.values_list('name', flat=True)))
            setattr(profile.user, '__dssoa__apigateway_update', False)

            return consumer

        elif status_code == 409:
            _status_code, consumer = get_apigw_consumer(consumer_id, profile.sso_id)

            setattr(profile.user, '__dssoa__apigateway_update', True)
            # setting apigw_consumer_id
            profile.apigw_consumer_id = consumer['id']
            # save profile
            profile.save()
            # add consumer to profile groups
            add_apigw_consumer_to_groups_acls(consumer['id'],
                                              app_settings.DEFAULT_PROFILE_GROUPS +
                                              list(profile.groups.values_list('name', flat=True)))
            setattr(profile.user, '__dssoa__apigateway_update', False)

            return consumer

        else:
            logger.error('Error ({0}) creating consumer for "{1}'.format(status_code, profile))
            raise Exception('Error ({0}) creating consumer for "{1}"'.format(status_code, profile))

    else:
        logger.error('Error ({0}) getting consumer for "{1}: {1}'.format(status_code, profile))
        raise Exception('Error ({0}) getting consumer for "{1}"'.format(status_code, profile))


def get_profile_apigw_consumer_id(profile, force_recreate=False):
    return get_or_create_profile_apigw_consumer(profile, force_recreate)['id']



import uuid as uuid_lib

import responses
from dateutil import tz

from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.test import TestCase, Client

from allauth.account.adapter import get_adapter

import jwt
from http.cookies import SimpleCookie
from rest_framework.authtoken.models import Token

from django_sso_app.core.apps.emails.models import EmailAddress
from django_sso_app.core.apps.services.models import Service, Subscription
from django_sso_app.core.apps.groups.models import Group
from django_sso_app.core.apps.devices.utils import add_profile_device

import string
import random
from datetime import datetime

from faker import Faker

fake = Faker()
User = get_user_model()


class UserTestCase(TestCase):
    def _get_random_uuid(self):
        return str(uuid_lib.uuid4())

    def _get_random_string(self, length=20):
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(length))

    def _get_random_pass(self, length=8):
        return self._get_random_string(length)

    def _get_random_username(self):
        return fake.user_name()

    def _get_random_email(self):
        return fake.ascii_safe_email()

    def _get_random_datetime(self, start_datetime=None):
        d = fake.date_time_between(start_date=start_datetime, end_date='+100y', tzinfo=tz.gettz('UTC'))
        return d.isoformat()

    def _get_random_birthdate(self, start_date='-99y', end_date='-18y'):
        d = fake.date_time_between(start_date=start_date, end_date=end_date)
        return d.strftime('%Y-%m-%d')

    def _get_client(self):
        return Client()

    def _get_allauth_adapter(self, request=None):
        return get_adapter(request)

    def _get_login_object(self, login, password):
        return {
            'login': login,
            'password': password,
            'fingerprint': self._get_random_string()
        }

    def _get_random_languagecode(self):
        L_CODES = ['en', 'it', 'fr', 'de']

        return random.choice(L_CODES)

    def _get_random_country(self):
        return self._get_random_languagecode().upper()

    def _get_random_latitude(self):
        return random.random() / 100

    def _get_random_longitude(self):
        return random.random() / 100

    def _get_new_profile_object(self, country='en', language='en', incomplete=False):
        from django_sso_app import app_settings

        obj = {}
        for f in app_settings.PROFILE_FIELDS:
            if f == 'birthdate':
                obj[f] = self._get_random_birthdate()
            elif f == 'expiration_date':
                obj[f] = self._get_random_datetime(start_datetime=datetime.now())
            elif f == 'picture':
                pass
            elif f == 'language':
                obj[f] = language or self._get_random_languagecode()
            elif f in ['country', 'country_of_birth']:
                obj[f] = country or self._get_random_country()
            elif f == 'latitude':
                obj[f] = self._get_random_latitude()
            elif f == 'longitude':
                obj[f] = self._get_random_longitude()
            elif f == 'gender':
                obj[f] = self._get_random_string(15)
            else:
                obj[f] = self._get_random_string()

        if incomplete:
            for f in app_settings.REQUIRED_PROFILE_FIELDS:
                obj[f] = None

        return obj

    def _get_new_profile_complete_user_object(self, username=None):
        return {
            'username': username or self._get_random_username(),
            'profile': self._get_new_profile_object()
        }

    def _get_new_invalid_profile_complete_user_object(self, username=None):
        from django_sso_app import app_settings

        profile_complete_object = self._get_new_profile_complete_user_object(username)
        profile_complete_object['profile'].pop(app_settings.REQUIRED_PROFILE_FIELDS[0])

        return profile_complete_object

    def _get_new_user_object(self, username=None, email=None, password=None, profile_object=None, incomplete=False):
        return {
            'username': username or self._get_random_username(),
            'email': email or self._get_random_email(),
            'password': password or self._get_random_pass(),
            'profile': profile_object or self._get_new_profile_object(incomplete=incomplete)
        }

    def _get_new_subscription_object(self, service_name=None, is_unsubscribed=False):
        domain_name = service_name or fake.domain_name()

        return {
            'service_url': 'http://{}'.format(domain_name),
            'is_unsubscribed': is_unsubscribed
        }

    def _get_new_group(self, name=None):

        group_name = name or self._get_random_string()

        return Group.objects.create(name=group_name)

    def _get_remote_user_object(self, uuid=None, username=None, email=None, password=None,
                                profile_object=None,
                                service_name=None,
                                service_subscribed=True,
                                group_name=None,
                                incomplete=False):
        from django_sso_app import app_settings

        user_uuid = uuid or self._get_random_uuid()
        user_object = self._get_new_user_object(username, email, password, profile_object, incomplete=incomplete)

        user_object['password'] = make_password(user_object['password'])

        user_object['sso_id'] = user_uuid
        user_object['sso_rev'] = random.randint(1, 100)

        user_object['url'] = app_settings.REMOTE_USER_URL.format(sso_id=user_uuid)

        user_object_profile_service = self._get_new_subscription_object(service_name, is_unsubscribed=not service_subscribed)

        user_object['profile']['subscriptions'] = [
            user_object_profile_service
        ]

        user_object['profile']['groups'] = [
            group_name or self._get_random_string()
        ]

        if incomplete:
            user_object['profile']['is_incomplete'] = True
            user_object['profile']['groups'] += [
                'incomplete'
            ]
            user_object['profile']['is_incomplete'] = True

        return user_object

    def _get_signup_object(self, username=None, email=None, referer=None, invalid=False):
        password = self._get_random_pass()

        obj = {
            'email': email or self._get_random_email(),
            'password1': password,
            'password2': password
        }

        if username is not None:
            obj['username'] = username

        if referer is not None:
            obj['referer'] = referer

        if invalid:
            obj.pop('email')

        return obj

    def _get_username_signup_object(self, username=None):
        return self._get_signup_object(username=username or self._get_random_username())

    def _get_invalid_signup_object(self):
        return self._get_signup_object(invalid=True)

    def _get_subscribe_signup_object(self, referer):
        return self._get_signup_object(referer=referer)

    def _get_new_service(self, name=None):
        domain_name = name or fake.domain_name()

        return Service.objects.create(service_url='http://{}'.format(domain_name), name=domain_name)

    def _get_new_user(self, username=None, email=None, password=None, service=None,
                      is_staff=False, is_superuser=False, incomplete=False):

        user = User(
            username=username or self._get_random_username(),
            email=email or self._get_random_email(),
            is_staff=is_staff,
            is_superuser=is_superuser
        )

        user.set_password(password or self._get_random_pass())

        if not is_staff and not is_superuser:
            _profile = self._get_new_profile_object(incomplete=incomplete)

            setattr(user, '__dssoa__profile__object', _profile)

            user.save()

            _user_email = EmailAddress.objects.create(user=user,
                                                      email=user.email,
                                                      primary=True,
                                                      verified=True)
        else:
            user.save()

        if service is not None:
            Subscription.objects.get_or_create(profile=user.sso_app_profile, service=service)

        return user

    def _get_new_staff_user(self, username=None, email=None, password=None):
        return self._get_new_user(username=username, email=email, password=password, is_staff=True)

    def _get_new_admin_user(self, username=None, email=None, password=None):
        return self._get_new_user(username=username, email=email, password=password, is_staff=True, is_superuser=True)

    def _get_new_incomplete_user(self, username=None, email=None, password=None):
        return self._get_new_user(username=username, email=email, password=password, incomplete=True)

    def _get_new_subscribed_user(self, service=None, username=None, email=None, password=None):
        subscribed_service = service or self._get_new_service()
        return self._get_new_user(service=subscribed_service, username=username, email=email, password=password)

    def _get_response_jwt_cookie(self, response):
        cookies = response.client.cookies.items()

        for (key, val) in cookies:
            if key == "jwt":
                return getattr(val, '_value')

    def _get_user_device(self, user, fingerprint=None):
        fp = fingerprint or self._get_random_string()

        return add_profile_device(user.sso_app_profile, fp)

    def _get_jwt(self, device, secret):
        from django_sso_app import app_settings

        secret = secret or device.apigw_jwt_secret

        return jwt.encode(
            device.get_jwt_payload(),
            secret,
            app_settings.JWT_ALGORITHM).decode('utf-8')

    def _get_jwt_cookie(self, device, secret=None):
        from django_sso_app import app_settings

        secret = secret or device.apigw_jwt_secret

        return SimpleCookie({
            app_settings.JWT_COOKIE_NAME: self._get_jwt(device, secret)
        })

    def _get_jwt_payload(self, sso_id=None, sso_rev=None, fingerprint=None):
        return {
            'id': random.randint(1, 1000),
            'fp': fingerprint or self._get_random_string(),
            'fingerprint': self._get_random_string(),
            'sso_id': sso_id or self._get_random_uuid(),
            'sso_rev': sso_rev or random.randint(0, 100)
        }

    def _get_valid_jwt(self, sso_id=None, sso_rev=None, fingerprint=None, secret=None):
        from django_sso_app import app_settings

        secret = secret or app_settings.TOKENS_JWT_SECRET

        valid_new_profile_jwt_payload = self._get_jwt_payload(sso_id, sso_rev, fingerprint)

        return jwt.encode(
            valid_new_profile_jwt_payload,
            secret,
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')

    def _get_valid_jwt_cookie(self, sso_id=None, sso_rev=None, fingerprint=None, secret=None):
        from django_sso_app import app_settings

        return SimpleCookie({
            app_settings.JWT_COOKIE_NAME: self._get_valid_jwt(sso_id, sso_rev, fingerprint, secret=secret)
        })

    def _get_new_api_token(self, user):
        return Token.objects.create(user=user)

    def _get_new_api_token_headers(self, user):
        token = self._get_new_api_token(user)

        return {
            'content_type': 'application/json',
            'HTTP_AUTHORIZATION': 'Token {}'.format(token.key)
        }

    def _set_mocked_response(self, mocked_url, mocked_response, status=200):
        responses.add(responses.GET, mocked_url,
                      json=mocked_response, status=status)

    def _test_response_location(self, response, location):
        self.assertEqual(response['location'], location)

    def _test_response_path(self, response, path):
        self.assertEqual(response.request['PATH_INFO'], path)



import os
from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "django_sso_app.config.settings.local")

app = Celery("django_sso_app")

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("django.conf:settings", namespace="CELERY")

# Load task modules from all registered Django app configs.
app.autodiscover_tasks()



from rest_framework import viewsets
from rest_framework import permissions
from rest_framework.response import Response

from .serializers import GroupSerializer, GroupingSerializer, GroupingListSerializer
from .models import Group, Grouping


class GroupApiViewSet(viewsets.ReadOnlyModelViewSet):

    serializer_class = GroupSerializer
    queryset = Group.objects.all()
    lookup_field = 'pk'
    permission_classes = (permissions.IsAuthenticated, )

    def list(self, request, *args, **kwargs):
        """
        List groups
        """
        return super(GroupApiViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, pk=None, *args, **kwargs):
        """
        Group detail
        """
        return super(GroupApiViewSet, self).retrieve(request, pk, *args, **kwargs)


class GroupingApiViewSet(viewsets.ReadOnlyModelViewSet):

    queryset = Grouping.get_root_nodes()
    lookup_field = 'pk'
    permission_classes = (permissions.IsAuthenticated, )

    def get_serializer_class(self):
        if self.action == 'list':
            return GroupingListSerializer
        else:
            return GroupingSerializer

    def list(self, request, *args, **kwargs):
        """
        List groupings
        """
        return super(GroupingApiViewSet, self).list(request, *args, **kwargs)

    def retrieve(self, request, pk=None, *args, **kwargs):
        """
        Grouping detail
        """

        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)



import os

from django.apps import apps
from django.template import Origin, TemplateDoesNotExist
from django.template.loaders.filesystem import Loader as BaseLoader


class Loader(BaseLoader):
    is_usable = True

    def get_template_sources(self, tpl):
        template_parts = tpl.split(":", 1)

        if len(template_parts) != 2:
            raise TemplateDoesNotExist(tpl)

        app_name, template_name = template_parts
        app = apps.get_app_config(app_name)
        template_dir = os.path.abspath(os.path.join(app.path, 'templates'))
        path = os.path.join(template_dir, template_name)

        yield Origin(
            name=path,
            template_name=tpl,
            loader=self,
        )



from django.conf.urls import url
from django.contrib.auth.decorators import login_required

from .views import SignupView, LoginView, LogoutView, EmailView, PasswordResetView, PasswordResetFromKeyView, \
                   PasswordChangeView, PasswordSetView


allauth_api_urlpatterns = [
    # allauth
    url(r"^api/v1/auth/signup/$", SignupView.as_view(), name="rest_signup"),
    url(r"^api/v1/auth/login/$", LoginView.as_view(), name="rest_login"),
    url(r"^api/v1/auth/logout/$", LogoutView.as_view(), name="rest_logout"),

    # E-mail
    url(r"^api/v1/auth/email/$", login_required(EmailView.as_view()), name="rest_email"),

    # Password
    url(r"^api/v1/auth/password/reset/$", PasswordResetView.as_view(), name="rest_password_reset"),
    url(r"^api/v1/auth/password/reset/key/$", PasswordResetFromKeyView.as_view(),
        name="rest_reset_password_from_key"),

    url(r"^api/v1/auth/password/change/$", PasswordChangeView.as_view(),
        name="rest_password_change"),

    url(r"^api/v1/auth/password/set/$", PasswordSetView.as_view(),
        name="rest_password_set"),
]



import logging

from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import get_user_model

from rest_framework import authentication

from .. import app_settings
from ..exceptions import AuthenticationFailed, InvalidToken
from .utils import get_request_jwt_header, jwt_decode

User = get_user_model()

logger = logging.getLogger('django_sso_app.core')


class JWTAuthentication(authentication.BaseAuthentication):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    """
    www_authenticate_realm = 'api'

    def authenticate(self, request):
        logger.info('Authenticating')

        header = self.get_header(request)
        if header is None:
            return None

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        validated_token = self.get_validated_token(raw_token)

        return self.get_user(validated_token), None

    def authenticate_header(self, request):
        return '{0} realm="{1}"'.format(
            app_settings.AUTH_HEADER_TYPES[0],
            self.www_authenticate_realm,
        )

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        return get_request_jwt_header(request, encoded=True)

    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()
        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if parts[0] not in app_settings.AUTH_HEADER_TYPE_BYTES:
            # Assume the header does not contain a JSON web token
            return None
        if len(parts) != 2:
            raise AuthenticationFailed(
                _('Authorization header must contain two space-delimited values'),
                code='bad_authorization_header',
            )

        return parts[1]

    def get_validated_token(self, raw_token):
        """
        Validates an encoded JSON web token and returns a validated token
        wrapper object.
        """
        messages = []

        try:
            _device, decoded_jwt = jwt_decode(raw_token, verify=True)

            return decoded_jwt

        except Exception as e:
            logger.info('error deconding token')
            messages.append('{}'.format(e))

        raise InvalidToken({
            'detail': _('Given token not valid for any token type'),
            'messages': messages,
        })

    def get_user(self, validated_token):
        """
        Attempts to find and return a user using the given validated token.
        """
        try:
            sso_id = validated_token[app_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_('Token contained no recognizable user identification'))

        try:
            user = User.objects.get(**{app_settings.USER_ID_FIELD: sso_id})
        except User.DoesNotExist:
            raise AuthenticationFailed(_('User not found'), code='user_not_found')

        if not user.is_active:
            raise AuthenticationFailed(_('User is inactive'), code='user_inactive')

        return user



import logging

from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import smart_str

from jwt.exceptions import InvalidSignatureError

from ... import app_settings
from ...utils import set_session_key, get_session_key, invalidate_cookie
from ...permissions import is_authenticated, is_django_staff
from ...exceptions import RequestHasValidJwtWithNoDeviceAssociated, ServiceSubscriptionRequiredException, ProfileIncompleteException
from ...tokens.utils import get_request_jwt, jwt_decode

from .backend import DjangoSsoAppAuthenticationBackendMiddleware
from .app import DjangoSsoAppAuthenticationAppMiddleware

logger = logging.getLogger('django_sso_app.core.authentication.middleware')

SSO_ID_JWT_KEY = 'sso_id'
FINGERPRINT_JWT_KEY = 'fp'


class DjangoSsoAppAuthenticationMiddleware(DjangoSsoAppAuthenticationBackendMiddleware,
                                           DjangoSsoAppAuthenticationAppMiddleware):
    """
    See django.contrib.auth.middleware.RemoteUserMiddleware.
    """

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "SSO middleware requires the authentication middleware to be"
                " installed.  Edit your MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the SsoMiddleware class.")

        request_path = request.path
        request_method = request.method
        request_ip = request.META.get('REMOTE_ADDR', None)
        requesting_user = request.user

        logger.info('--> "{}" request "{}" path "{}" method "{}" user "{}"'.format(request_ip,
                                                                                   id(request),
                                                                                   request_path,
                                                                                   request_method,
                                                                                   requesting_user))

        if is_authenticated(requesting_user):

            if is_django_staff(requesting_user):
                logger.info('Skipping django staff user "{}"'.format(requesting_user))
                #  >= 1.10 has is_authenticated as parameter
                # If a staff user is already authenticated, we don't need to
                # continue

                return

            elif self._is_admin_path(request):
                logger.warning('Non staff user "{}" called admin path'.format(requesting_user))

                self._remove_invalid_user(request)
                self._clear_response_jwt(request)

                return

        # saving request info
        set_session_key(request, '__dssoa__request_ip', request_ip)
        set_session_key(request, '__dssoa__requesting_user', requesting_user)

        """
        if self.request_path.startswith('/api/v1/passepartout'):
            logger.info('is passepartout path')
            # set_session_key(request, '__dssoa__is_passepartout_path', True)
        """

        try:
            request_jwt = get_request_jwt(request, encoded=False)

            if request_jwt is None:
                logger.info('No JWT in request, skipping')

                return

            else:
                # decoding request JWT
                request_device, decoded_jwt = jwt_decode(request_jwt, verify=True)

        except KeyError:
            logger.exception('Malformed JWT "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except InvalidSignatureError:
            logger.warning('Invalid JWT signature "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except RequestHasValidJwtWithNoDeviceAssociated:
            logger.warning('RequestHasValidJwtWithNoDeviceAssociated "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        except Exception:
            logger.exception('Generic middleware exception "{}"'.format(request_jwt))

            self._remove_invalid_user(request)
            self._clear_response_jwt(request)

            return

        else:
            # caching device fingerprint
            request_device_fingerprint = decoded_jwt[FINGERPRINT_JWT_KEY]
            logger.debug('Caching request fingerprint: {}'.format(request_device_fingerprint))
            set_session_key(request, '__dssoa__device__fingerprint', request_device_fingerprint)

        apigateway_enabled = app_settings.APIGATEWAY_ENABLED
        request_jwt_sso_id = decoded_jwt[SSO_ID_JWT_KEY]

        if apigateway_enabled:
            consumer_custom_id = request.META.get(self.consumer_id_header, None)
            consumer_is_anonymous_header = request.META.get(self.anonymous_consumer_header, None)

            if consumer_is_anonymous_header is not None and \
                    consumer_is_anonymous_header == self.anonymous_consumer_header_value:
                consumer_is_anonymous = True
            else:
                if consumer_custom_id is None:
                    consumer_is_anonymous = True
                else:
                    if consumer_custom_id in self.anonymous_consumer_custom_ids:
                        consumer_is_anonymous = True
                    else:
                        consumer_is_anonymous = False

            if consumer_is_anonymous:
                # If anonymous consumer header specified or consumer custom id header don't exist
                # or consumer custom_id is one of anonymous consumer custom_ids
                # then remove any existing authenticated user and return (leaving request.user set to
                # AnonymousUser by the AuthenticationMiddleware).

                logger.info('User is anonymous')

                self._remove_invalid_user(request)

                return

            else:
                sso_id = consumer_custom_id

                set_session_key(request, '__dssoa__apigateway__consumer_custom_id', consumer_custom_id)

        else:
            sso_id = request_jwt_sso_id

        # If the user is already authenticated and that user is active and the
        # one we are getting passed in the headers, then the correct user is
        # already persisted in the session and we don't need to continue.
        if is_authenticated(requesting_user):
            # double check sso_id equality between received jwt and sso_app_profile
            if smart_str(requesting_user.sso_id) != sso_id:

                # An authenticated user is associated with the request, but
                # it does not match the authorized user in the header.
                logger.warning('credentials and request_user sso_id differs! "{}" "{}"'.format(sso_id,
                                                                                               requesting_user.sso_id))

                self._remove_invalid_user(request)
                self._clear_response_jwt(request)

                return

        try:
            # authentication
            if app_settings.BACKEND_ENABLED:
                self.backend_process_request(request, request_jwt, decoded_jwt)
            else:
                self.app_process_request(request, request_jwt, decoded_jwt)

        except ProfileIncompleteException as e:
            if self._request_path_is_disabled_for_incomplete_users(request):
                logger.info('User must complete profile')
                set_session_key(request, '__dssoa__redirect', e.response)

        except ServiceSubscriptionRequiredException as e:
            if self._request_path_is_disabled_for_users_to_subscribe(request):
                logger.info('User must subscribe service')
                set_session_key(request, '__dssoa__redirect', e.response)

        except Exception as e:
            logger.exception('Generic middleware backend exception "{}"'.format(e))

            self._remove_invalid_user(request)

            return

    @staticmethod
    def _process_response(request, response):
        redirect = get_session_key(request, '__dssoa__redirect', None)

        if redirect is not None:
            return redirect

        # invalidate JWT cookie on response (if required)
        if get_session_key(request, '__dssoa__clear_response_jwt', False):
            invalidate_cookie(response, app_settings.JWT_COOKIE_NAME)

        return response

    def process_response(self, request, response):
        # getting request info
        requesting_user = getattr(request, 'user', get_session_key(request, '__dssoa__requesting_user', None))
        request_ip = get_session_key(request, '__dssoa__request_ip')

        user_logged_in = get_session_key(request, '__dssoa__logged_in', False)  # !!
        user_logged_out = get_session_key(request, '__dssoa__logged_out', False)
        request_fp = get_session_key(request, '__dssoa__device__fingerprint', None)

        logger.debug('login: {} - logout: {} - FP: {}'.format(user_logged_in, user_logged_out, request_fp))

        response = self._process_response(request, response)

        logger.info('<-- "{}" request "{}" user "{}" path "{}" method "{}" ({})'.format(request_ip,
                                                                                        id(request),
                                                                                        requesting_user,
                                                                                        request.path,
                                                                                        request.method,
                                                                                        response.status_code))

        return response



import logging
import uuid

from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

from django_countries.fields import CountryField

from ..groups.models import Group
from ... import app_settings
from ...models import CreatedAtModel, UpdatableModel, DeactivableModel, PublicableModel

logger = logging.getLogger('django_sso_app.core.apps.profiles')


def _get_new_uuid():
    return str(uuid.uuid4())


class SsoIdPKManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, sso_id):
        return self.get(sso_id=sso_id)


class AbstractProfileModel(CreatedAtModel, UpdatableModel, DeactivableModel, PublicableModel):
    """
    Profile abstract model

    """

    class Meta:
        abstract = True

    objects = SsoIdPKManager()

    def natural_key(self):
        return (self.sso_id, )

    sso_id = models.CharField(_('sso id'), max_length=36, unique=True, default=_get_new_uuid, db_index=True)  # unique on not null , editable=False
    sso_rev = models.PositiveIntegerField(_('sso revision'), default=1)

    django_user_username = models.CharField(_('user username'), max_length=150, null=True, blank=True, editable=False)
    django_user_email = models.EmailField(_('email address'), null=True, blank=True, editable=False)

    external_id = models.CharField(_('external id'), max_length=36, null=True, blank=True, db_index=True)

    ssn = models.CharField(_('social security number'), max_length=255, null=True, blank=True)
    phone = models.CharField(_('phone'), max_length=255, null=True, blank=True)

    first_name = models.CharField(_('first name'), max_length=255, null=True,
                                  blank=True)
    last_name = models.CharField(_('last name'), max_length=255, null=True,
                                 blank=True)
    alias = models.CharField(_('alias'), max_length=255, null=True, blank=True)

    description = models.TextField(_('description'), null=True, blank=True)
    picture = models.TextField(_('picture'), null=True, blank=True)

    birthdate = models.DateField(_('birthdate'), null=True, blank=True)
    country_of_birth = CountryField(_('country of birth'), max_length=46, null=True, blank=True)

    latitude = models.FloatField(_('latitude'), null=True, blank=True)
    longitude = models.FloatField(_('longitude'), null=True, blank=True)

    # country = models.CharField(_('country'), max_length=46, null=True,
    #                            blank=True)
    country = CountryField(_('country'), max_length=46, null=True, blank=True)

    address = models.TextField(_('address'), null=True, blank=True)
    language = models.CharField(_('language'), max_length=3, null=True, blank=True)

    gender = models.CharField(_('gender'), max_length=15, null=True, blank=True)

    company_name = CountryField(_('company name'), max_length=255, null=True, blank=True)

    expiration_date = models.DateTimeField(_('expiration date'), null=True, blank=True)

    completed_at = models.DateTimeField(_('profile completion date'), null=True, blank=True)

    unsubscribed_at = models.DateTimeField(_('unsubscription date'), null=True, blank=True)

    meta = models.TextField(_('meta'), null=True, blank=True)

    @property
    def is_unsubscribed(self):
        return self.unsubscribed_at is not None

    @is_unsubscribed.setter
    def is_unsubscribed(self, value):
        self.unsubscribed_at = timezone.now()


class Profile(AbstractProfileModel):
    class Meta:
        app_label = 'django_sso_app'
        verbose_name = _('Profile')

    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
                                null=True, blank=True,
                                verbose_name=_("user"),
                                related_name="sso_app_profile")

    apigw_consumer_id = models.CharField(_('API gateway consumer id'), max_length=36, null=True, blank=True)

    groups = models.ManyToManyField(Group, blank=True, related_name='sso_app_profiles', verbose_name=_('groups'),
                                    related_query_name='sso_app_profile')

    def __str__(self):
        return '{}:{}:{}'.format(self.email, self.sso_id, self.username)

    def update_rev(self, commit=False):
        if getattr(self, '__rev_updated', False):
            logger.debug('Rev already updated')
        else:
            logger.info(
                'Updating rev for User {}, from {} to {}, commit? {}'.format(self,
                                                                             self.sso_rev,
                                                                             self.sso_rev + 1,
                                                                             commit))

            self.sso_rev = models.F('sso_rev') + 1

            setattr(self, '__rev_updated', True)

        if commit:
            self.save()

    def save(self, *args, **kwargs):
        # pre save
        creating = self._state.adding

        # set profile completed_at if all required fields are set
        if not self.is_incomplete and self.completed_at is None:
            self.completed_at = timezone.now()

        if not creating:
            user = self.user

            if getattr(user, '__dssoa__apigateway_update', False):
                logger.debug('Profile updated by apigateway, skipping rev update')
            elif getattr(user, '__dssoa__creating', False):
                logger.debug('Created, skipping rev update')
            else:
                logger.debug('Profile model updated, updating rev')
                self.update_rev(False)

        super(AbstractProfileModel, self).save(*args, **kwargs)

        # post save
        # refresh instance from db because of rev_update
        if not creating:
            self.refresh_from_db()

    def get_relative_rest_url(self):
        return reverse("django_sso_app_profile:rest-detail", args=[self.sso_id])

    def add_to_group(self, group_name, creating=False):
        logger.info('Adding "{}" to group "{}"'.format(self, group_name))

        g, _created = Group.objects.get_or_create(name=group_name)
        if _created:
            logger.info('New group "{}" created'.format(g))

        if creating:
            setattr(self.user, '__dssoa__creating', True)

        if g not in self.groups.all():
            self.groups.add(g)

        if creating:
            setattr(self.user, '__dssoa__creating', False)

    def remove_from_group(self, group_name, creating=False):
        logger.info('Removing "{}" from group "{}"'.format(self, group_name))

        g, _created = Group.objects.get_or_create(name=group_name)
        if _created:
            logger.info('New group "{}" created'.format(g))

        if creating:
            setattr(self.user, '__dssoa__creating', True)

        if g in self.groups.all():
            self.groups.remove(g)

        if creating:
            setattr(self.user, '__dssoa__creating', False)

    @property
    def username(self):
        if getattr(self, 'user', None) is not None:
            return self.user.username
        else:
            return self.django_user_username

    @property
    def email(self):
        if getattr(self, 'user', None) is not None:
            return self.user.email
        else:
            return self.django_user_email

    @property
    def is_incomplete(self):
        for field in app_settings.REQUIRED_PROFILE_FIELDS:
            if getattr(self, field, None) in (None, ''):
                logger.debug('Empty "{}" field'.format(field))
                return True

        return False

    @property
    def must_subscribe(self):
        if app_settings.SERVICE_SUBSCRIPTION_REQUIRED:
            user_service_subscription = self.subscriptions.filter(service__service_url=app_settings.SERVICE_URL).first()

            if user_service_subscription is None or user_service_subscription.is_unsubscribed:
                return True

        return self.is_unsubscribed


"""
class DjangoSsoAppUserStatus(object):
    def __init__(self, is_active, is_unsubscribed, is_to_subscribe):
        self.is_active = is_active
        self.is_unsubscribed = is_unsubscribed
        self.is_to_subscribe = is_to_subscribe

    def __repr__(self):
        return ("{}(is_active={}, is_unsubscribed={}, is_to_subscribe={})"
                .format(self.__class__.__name__, self.is_active,
                        self.is_unsubscribed, self.is_to_subscribe))

    @staticmethod
    def get_user_status(is_unsubscribed, subscriptions, email_verified=True):
        _is_active = app_settings.DEACTIVATE_USER_ON_UNSUBSCRIPTION

        if is_unsubscribed:  # is unsubscribed from sso
            return DjangoSsoAppUserStatus(is_active=_is_active,
                                          is_unsubscribed=True,
                                          is_to_subscribe=False)
        else:  # is still subscribed to sso
            for subscription in subscriptions:
                if subscription["service_url"] == app_settings.APP_URL:  #
                    # subscription found
                    if subscription["is_unsubscribed"]:  # user is unsubscribed from service
                        return DjangoSsoAppUserStatus(is_active=(_is_active
                                                                 and email_verified),
                                                      is_unsubscribed=True,
                                                      is_to_subscribe=False)
                    return DjangoSsoAppUserStatus(is_active=email_verified,
                                                  is_unsubscribed=False,
                                                  is_to_subscribe=False)

            # is NOT subscribed
            return DjangoSsoAppUserStatus(
                is_active=email_verified, is_unsubscribed=False,
                is_to_subscribe=app_settings.SERVICE_SUBSCRIPTION_IS_MANDATORY)

"""



# Generated by Django 2.2.9 on 2020-02-22 13:25

from django.conf import settings
import django.contrib.auth.models
from django.db import migrations, models
import django.db.models.deletion
import django_countries.fields
import django_sso_app.core.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('is_public', models.BooleanField(default=False, verbose_name='is public')),
                ('ssn', models.CharField(blank=True, max_length=255, null=True, verbose_name='social security number')),
                ('phone', models.CharField(blank=True, max_length=255, null=True, verbose_name='phone')),
                ('first_name', models.CharField(blank=True, max_length=255, null=True, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=255, null=True, verbose_name='last name')),
                ('description', models.TextField(blank=True, null=True, verbose_name='description')),
                ('picture', models.TextField(blank=True, null=True, verbose_name='picture')),
                ('birthdate', models.DateField(blank=True, null=True, verbose_name='birthdate')),
                ('latitude', models.FloatField(blank=True, null=True, verbose_name='latitude')),
                ('longitude', models.FloatField(blank=True, null=True, verbose_name='longitude')),
                ('country', django_countries.fields.CountryField(blank=True, max_length=2, null=True, verbose_name='country')),
                ('address', models.TextField(blank=True, null=True, verbose_name='address')),
                ('language', models.CharField(blank=True, max_length=3, null=True, verbose_name='language')),
                ('sso_id', models.CharField(db_index=True, default=django_sso_app.core.apps.profiles.models._get_new_uuid, max_length=36, unique=True, verbose_name='sso id')),
                ('sso_rev', models.PositiveIntegerField(default=1, verbose_name='sso revision')),
                ('apigw_consumer_id', models.CharField(blank=True, max_length=36, null=True, verbose_name='API gateway consumer id')),
                ('unsubscribed_at', models.DateTimeField(blank=True, null=True, verbose_name='unsubscription date')),
                ('user', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sso_app_profile', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'Profile',
            },
        ),
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('name', models.CharField(max_length=32, null=True, unique=True, verbose_name='name')),
                ('service_url', models.CharField(max_length=255, unique=True, verbose_name='service url')),
                ('picture', models.TextField(blank=True, null=True, verbose_name='picture')),
                ('cookie_domain', models.CharField(blank=True, max_length=255, null=True, verbose_name='cookie domain')),
                ('redirect_wait', models.PositiveIntegerField(default=2000, verbose_name='redirect wait')),
                ('subscription_required', models.BooleanField(default=False, verbose_name='subscription required')),
                ('is_public', models.BooleanField(default=True, verbose_name='is public')),
            ],
            options={
                'verbose_name': 'Service',
            },
        ),
        migrations.CreateModel(
            name='Status',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('shape', models.CharField(default='backend_only', max_length=255)),
            ],
            options={
                'verbose_name': 'Status',
            },
            bases=(django_sso_app.core.models.UuidPKModel, models.Model),
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
            ],
            options={
                'verbose_name': 'Group',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('auth.group',),
            managers=[
                ('objects', django.contrib.auth.models.GroupManager()),
            ],
        ),
        migrations.CreateModel(
            name='TOS',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('language', models.CharField(max_length=35, verbose_name='language')),
                ('text', models.TextField(default='', verbose_name='text')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='terms_of_service', to='django_sso_app.Service', verbose_name='service')),
            ],
            options={
                'verbose_name': 'TOS',
            },
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('is_public', models.BooleanField(default=False, verbose_name='is public')),
                ('unsubscribed_at', models.DateTimeField(blank=True, null=True, verbose_name='unsubscribed at')),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='django_sso_app.Profile', to_field='sso_id', verbose_name='profile')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='django_sso_app.Service', verbose_name='service')),
            ],
            options={
                'verbose_name': 'Subscription',
            },
        ),
        migrations.AddField(
            model_name='service',
            name='groups',
            field=models.ManyToManyField(blank=True, related_name='sso_app_services', to='django_sso_app.Group', verbose_name='groups'),
        ),
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('fingerprint', models.CharField(max_length=32, verbose_name='fingerprint')),
                ('user_agent', models.CharField(blank=True, max_length=32, null=True, verbose_name='user agent')),
                ('apigw_jwt_id', models.CharField(blank=True, max_length=36, null=True, verbose_name='apigw_jwt_id')),
                ('apigw_jwt_key', models.CharField(blank=True, max_length=32, null=True, verbose_name='apigw_jwt_key')),
                ('apigw_jwt_secret', models.CharField(blank=True, max_length=32, null=True, verbose_name='apigw_jwt_secret')),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='devices', to='django_sso_app.Profile', verbose_name='profile')),
            ],
            options={
                'verbose_name': 'Device',
            },
        ),
        migrations.CreateModel(
            name='Passepartout',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='updated at')),
                ('deleted', models.BooleanField(default=False, verbose_name='deleted')),
                ('token', models.CharField(db_index=True, max_length=36, unique=True, verbose_name='token')),
                ('jwt', models.TextField(blank=True, null=True, verbose_name='jwt')),
                ('bumps', models.PositiveIntegerField(default=0, verbose_name='bumps')),
                ('device', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_sso_app.Device', verbose_name='device')),
            ],
            options={
                'verbose_name': 'Passepartout',
                'unique_together': {('token', 'deleted')},
            },
        ),
    ]
